
import os
import torch
import torchvision
import numpy as np
import cv2
import pandas as pd
from PIL import Image

def segmentation(input_image, unet,device,image_name, fileext,save_path):
    """
        This function segments the lung region from the given input image
        
        inputs - 
            input_image: image input of type Image.PIL
            unet: lung segmentation model
            device: torch.device("cpu") ot torch.device("gpu")
            image_name: name of the image to save the segmented lung region
            filext: extension of the image to save. "png, jpg or jpeg".
            save_path: folder path to save the segmented image.
        
        returns nothing. saves the segmented lung region 
    """
    
    origin = torchvision.transforms.functional.resize(input_image, (512, 512))
    origin = torchvision.transforms.functional.to_tensor(origin) - 0.5
    with torch.no_grad():
        origin = torch.stack([origin])
        origin = origin.to(device)
        out = unet(origin) #logits
        softmax = torch.nn.functional.log_softmax(out, dim=1)
        out = torch.argmax(softmax, dim=1)
    origin = origin[0].to("cpu")
    lung_masked = out[0].to("cpu")
    
    mi = np.array(lung_masked, np.uint8)
    np_image = np.array(input_image.resize((512,512)), np.uint8)
    
    mi = (mi[:, :] * 255.).astype(np.uint8)

    resized = cv2.bitwise_and(np_image, mi)
    resized=resized*255
    resized=resized.astype('uint8')
    return resized


        
def segmentation_func(image_path, segmentation_model, device):
    
    """
        external function used to call the main segmentation function.
        input-
            image_path: path to the image.
            segmentation_model: lung segmentation model
            device: torch.device("cpu") ot torch.device("gpu")
    """
    
    full_path = image_path    
    if 1:
        input_image = Image.open(full_path).convert("P")
        
        trial_save_lung = full_path.rstrip(os.path.basename(full_path))
        
        _, fileext = os.path.splitext(os.path.basename(full_path))
        seg_image_name = 'segmented'
        
        
        resized = segmentation(input_image, segmentation_model, device, seg_image_name, fileext, trial_save_lung)
        
        
        return resized

        

class Block(torch.nn.Module):
    def __init__(self, in_channels, mid_channel, out_channels, batch_norm=False):
        super().__init__()
        
        self.conv1 = torch.nn.Conv2d(in_channels=in_channels, out_channels=mid_channel, kernel_size=3, padding=1)
        self.conv2 = torch.nn.Conv2d(in_channels=mid_channel, out_channels=out_channels, kernel_size=3, padding=1)
        
        self.batch_norm = batch_norm
        if batch_norm:
            self.bn1 = torch.nn.BatchNorm2d(mid_channel)
            self.bn2 = torch.nn.BatchNorm2d(out_channels)
            
    def forward(self, x):
        x = self.conv1(x)
        if self.batch_norm:
            x = self.bn1(x)
        x = torch.nn.functional.relu(x, inplace=True)
        
        x = self.conv2(x)
        if self.batch_norm:
            x = self.bn2(x)
        out = torch.nn.functional.relu(x, inplace=True)
        return out


class PretrainedUNet(torch.nn.Module):
    def up(self, x, size):
        return torch.nn.functional.interpolate(x, size=size, mode=self.upscale_mode)
    
    
    def down(self, x):
        return torch.nn.functional.max_pool2d(x, kernel_size=2)
    
    
    def __init__(self, in_channels, out_channels, batch_norm=False, upscale_mode="nearest"):
        super().__init__()
        
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.batch_norm = batch_norm
        self.upscale_mode = upscale_mode

        self.init_conv = torch.nn.Conv2d(in_channels, 3, 1)

        endcoder = torchvision.models.vgg11(pretrained=True).features
        self.conv1 = endcoder[0]   # 64
        self.conv2 = endcoder[3]   # 128
        self.conv3 = endcoder[6]   # 256
        self.conv3s = endcoder[8]  # 256
        self.conv4 = endcoder[11]   # 512
        self.conv4s = endcoder[13]  # 512
        self.conv5 = endcoder[16]  # 512
        self.conv5s = endcoder[18] # 512

        self.center = Block(512, 512, 256, batch_norm)

        self.dec5 = Block(512 + 256, 512, 256, batch_norm)
        self.dec4 = Block(512 + 256, 512, 128, batch_norm)
        self.dec3 = Block(256 + 128, 256, 64, batch_norm)
        self.dec2 = Block(128 + 64, 128, 32, batch_norm)
        self.dec1 = Block(64 + 32, 64, 32, batch_norm)

        self.out = torch.nn.Conv2d(in_channels=32, out_channels=out_channels, kernel_size=1)

    def forward(self, x): 
        init_conv = torch.nn.functional.relu(self.init_conv(x), inplace=True)

        enc1 = torch.nn.functional.relu(self.conv1(init_conv), inplace=True)
        enc2 = torch.nn.functional.relu(self.conv2(self.down(enc1)), inplace=True)
        enc3 = torch.nn.functional.relu(self.conv3(self.down(enc2)), inplace=True)
        enc3 = torch.nn.functional.relu(self.conv3s(enc3), inplace=True)
        enc4 = torch.nn.functional.relu(self.conv4(self.down(enc3)), inplace=True)
        enc4 = torch.nn.functional.relu(self.conv4s(enc4), inplace=True)
        enc5 = torch.nn.functional.relu(self.conv5(self.down(enc4)), inplace=True)
        enc5 = torch.nn.functional.relu(self.conv5s(enc5), inplace=True)

        center = self.center(self.down(enc5))

        dec5 = self.dec5(torch.cat([self.up(center, enc5.size()[-2:]), enc5], 1))
        dec4 = self.dec4(torch.cat([self.up(dec5, enc4.size()[-2:]), enc4], 1))
        dec3 = self.dec3(torch.cat([self.up(dec4, enc3.size()[-2:]), enc3], 1))
        dec2 = self.dec2(torch.cat([self.up(dec3, enc2.size()[-2:]), enc2], 1))
        dec1 = self.dec1(torch.cat([self.up(dec2, enc1.size()[-2:]), enc1], 1))

        out = self.out(dec1)

        return out


import pandas as pd
import warnings

warnings.filterwarnings('ignore')

def test():
    print("Sahil")
    return

def _calculate_single_iou(bb1, bb2):
    """
    Calculate the Intersection over Union (IoU) of two bounding boxes.

    Parameters
    ----------
    bb1 : dict
        Keys: {'x1', 'y1', 'x2', 'y2'}
        Keys: {xMinCorr, yMaxCorr, xMaxCorr, yMinCorr}
        The (x1, y1) position is at the bottom left corner,
        the (x2, y2) position is at the top right corner
    bb2 : dict
        Keys: {'x1', 'y1', 'x2', 'y2'}
        The (x1, y1) position is at the bottom left corner,
        the (x2, y2) position is at the top right corner

    Returns
    -------
    float
        in [0, 1]
    """
    
    if(bb2['x1']==0 and bb2['x2']==0 and bb2['y1']==0 and bb2['y2']==0):
        return 0.0
    assert bb1['x1'] < bb1['x2']
    assert bb1['y1'] < bb1['y2']
    assert bb2['x1'] < bb2['x2']
    assert bb2['y1'] < bb2['y2']

    # determine the coordinates of the intersection rectangle
    x_left = max(bb1['x1'], bb2['x1'])
    y_top = max(bb1['y1'], bb2['y1'])
    x_right = min(bb1['x2'], bb2['x2'])
    y_bottom = min(bb1['y2'], bb2['y2'])

    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes is always an
    # axis-aligned bounding box
    intersection_area = abs((x_right - x_left) * (y_bottom - y_top))

    # compute the area of both AABBs
    bb1_area = abs((bb1['x2'] - bb1['x1']) * (bb1['y2'] - bb1['y1']))
    bb2_area = abs((bb2['x2'] - bb2['x1']) * (bb2['y2'] - bb2['y1']))

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
    assert iou >= 0.0
    assert iou <= 1.0
    return iou

def li_score(df_pred, df_pred_wp, df_gt, iou_thresholds = [0.5]):
    
    df_iou_org = pd.DataFrame(columns=iou_thresholds)
    df_iou_org.columns = iou_thresholds
    
    for i in (range(len(df_gt))):
        name1 = df_gt['file_name'][i].split('_')[0]
        for j in range(len(df_pred)):
            name2 = df_pred['file_name'][j].split('_')[0]
            if name1==name2:
                # print(name1)
                bb1 = {'x1':df_gt.iloc[i]['x_min'], 'y1':df_gt.iloc[i]['y_min'], 'x2':df_gt.iloc[i]['x_max'],'y2': df_gt.iloc[i]['y_max']}
                bb2 = {'x1':df_pred.iloc[j]['x_min'], 'y1':df_pred.iloc[j]['y_min'], 'x2':df_pred.iloc[j]['x_max'],'y2': df_pred.iloc[j]['y_max']}
                
                if df_gt['class'][i] == df_pred['class'][j]:
                    score = _calculate_single_iou(bb1,bb2)

                else:
                    score = 0
                cnt_list = []
                
                for iou_threshold in iou_thresholds:
                    cnt = 0
                    if score>=iou_threshold:
                        cnt=1
                    cnt_list.append(cnt)
                # print(cnt_list)
                df_iou_org.loc[df_gt['file_name'][i]] = cnt_list
        #     break
        # break



    df_iou_wp = pd.DataFrame(columns=iou_thresholds)
    df_iou_wp.columns = iou_thresholds
    
    for i in (range(len(df_gt))):
        name1 = df_gt['file_name'][i].split('_')[0]
        for j in range(len(df_pred_wp)):
            name2 = df_pred_wp['file_name'][j].split('_')[0]
            if name1==name2:
                # print(name1)
                bb1 = {'x1':df_gt.iloc[i]['x_min'], 'y1':df_gt.iloc[i]['y_min'], 'x2':df_gt.iloc[i]['x_max'],'y2': df_gt.iloc[i]['y_max']}
                bb2 = {'x1':df_pred_wp.iloc[j]['x_min'], 'y1':df_pred_wp.iloc[j]['y_min'], 'x2':df_pred_wp.iloc[j]['x_max'],'y2': df_pred_wp.iloc[j]['y_max']}

                if df_gt['class'][i] == df_pred_wp['class'][j]:
                    score = _calculate_single_iou(bb1,bb2)
                    
                else:
                    score = 0

                cnt_list = []
                # cnt = 0
                for iou_threshold in iou_thresholds:
                    cnt = 0
                    if score>=iou_threshold:
                        cnt=1
                    cnt_list.append(cnt)
                # print(cnt_list)
                df_iou_wp.loc[df_gt['file_name'][i]] = cnt_list
        #     break
        # break
        
    df_iou_org = df_iou_org.reset_index()
    df_iou_wp = df_iou_wp.reset_index()
    
    
    df_pi = pd.DataFrame(columns=iou_thresholds)
    
    for i in (range(len(df_iou_org))):
        row_org = df_iou_org.iloc[[i]]
        row_wp = df_iou_wp.iloc[[i]]
        # print(row_org["Name"])
        # print(row_wp["Name"])
        cols = iou_thresholds
        ans = []
        for c in cols:
            # if(int(row_org[c]))
            if (int(row_org[c])>0 and int(row_wp[c])==0) or (int(row_org[c])==0 and int(row_wp[c])>0):
                ans.append(1)
            else:
                ans.append(0)
        
        df_pi.loc[df_iou_org.iloc[[i]]['index'].values [0]] = ans
        
    df_pi = df_pi.reset_index()
    
    unique_names = []
    for i in (range(len(df_pi))):
        if df_pi.iloc[i]['index'].split('_')[0] not in unique_names:
            unique_names.append(df_pi.iloc[i]['index'].split('_')[0])
    
    
    df_li = pd.DataFrame(columns=df_pi.columns)
    
    df_li = df_li.set_index('index')
    
    for i in (range(len(df_pi))):
        values = list(df_pi.iloc[i][iou_thresholds])
        if df_pi.iloc[i]['index'].split('_')[0] in df_li.index:
            # print(df2.loc[df.iloc[i]['index'].split('_')[0]] + values)
            df_li.loc[df_pi.iloc[i]['index'].split('_')[0]]  = df_li.loc[df_pi.iloc[i]['index'].split('_')[0]]  + values
        else:
            df_li.loc[df_pi.iloc[i]['index'].split('_')[0]]  = values
        
    for name in unique_names:
        org_values = list(df_li.loc[name])
        for i in range(len(org_values)):
            org_values[i] = int(org_values[i] > 0)
        df_li.loc[name] = org_values
        
    return df_li.mean().to_dict()


import os
import torch
from imageio import imread
from lung_segmentation import segmentation_func, PretrainedUNet
from skimage.transform import resize
import numpy as np
from tqdm import tqdm
import pandas as pd
import warnings

warnings.filterwarnings('ignore')

def load_segmentation_model(model_path, device):
    """
    Loading the lung segmentation Model
    
    inputs - 
        model_path: path to the .pt file
        device: 'cpu' or 'cuda'
        
    output -
        segmentation_model: lung segmentation model loaded on the specified device
    """
    segmentation_model = torch.load(model_path)
    segmentation_model.to(device)
    segmentation_model.eval()
    return segmentation_model


def get_dataframe(image_path, heatmap_hr_dir, heatmap_lr_dir, segmentation_model, device, image_size, thresholds):
    """
    Function to get dataframe of diceindex between hr and lr images
    
    Args:
        image_path (str): Path to the original images.
        heatmap_hr_dir (str): Directory of high-resolution heatmaps.
        heatmap_lr_dir (str): Directory of low-resolution heatmaps.
        segmentation_model: Loaded lung segmentation model.
        device: Torch device (CPU or CUDA).
        image_size (int): Size of the images.
        thresholds (list): List of thresholds.
        
    Returns:
        pd.DataFrame: DataFrame with calculated scores.
    """
    filenames = [f for f in os.listdir(image_path) if not f.startswith('.')]
    df = pd.DataFrame(columns=['Image_Name'] + [f'HR_Score_{t}' for t in thresholds] + [f'LR_Score_{t}' for t in thresholds])

    iter_filenames = tqdm(enumerate(filenames))

    for idx, fname in iter_filenames:
        row = {}
        lung_region = segmentation_func(os.path.join(image_path, fname), segmentation_model, device)
        lung_region = resize(lung_region, (image_size, image_size))
        
        heatmap_hr = imread(os.path.join(heatmap_hr_dir, fname))
        heatmap_lr = imread(os.path.join(heatmap_lr_dir, fname))
        row['Image_Name'] = fname
        
        for threshold in thresholds:
            dice_hr = (np.sum((heatmap_hr > int(threshold * 255)) * (lung_region > 0)) / 
                       (np.sum(heatmap_hr > int(threshold * 255)) + 1e-7))
            dice_lr = (np.sum((heatmap_lr > int(threshold * 255)) * (lung_region > 0)) / 
                       (np.sum(heatmap_lr > int(threshold * 255)) + 1e-7))
            row[f'HR_Score_{threshold}'] = dice_hr
            row[f'LR_Score_{threshold}'] = dice_lr

        df = df._append(row, ignore_index=True)
        
    return df


def ols_score(image_path, heatmap_hr_dir, heatmap_lr_dir, device='cpu', image_size=512, thresholds=[0.5]):
    """Function to compute OLS Score and return the results as a DataFrame
    
    Args:
        image_path (str): Path to the original images.
        output_dir (str): Directory to save the output CSV file.
        model_cam_name_hr (str): Directory for high-resolution heatmaps.
        model_cam_name_lr (str): Directory for low-resolution heatmaps.
        device (str): Device to use for computation ('cpu' or 'cuda').
        image_size (int): Size to resize the images.
        thresholds (list): List of thresholds for score calculation.
        
    Returns:
        pd.DataFrame: DataFrame containing the OLS scores.
    """
    
    model_path = 'lung_segmentation.pt'
    segmentation_model = load_segmentation_model(model_path, device)
    
    df = get_dataframe(image_path, heatmap_hr_dir, heatmap_lr_dir, segmentation_model, device, image_size, thresholds)
    
    return df


# Example usage
if __name__ == '__main__':
    image_path = '/home/dhaval/Templates/Sahil/OLS_Lib/images/original'
    heatmap_hr_dir = '/home/dhaval/Templates/Sahil/OLS_Lib/heat_maps_ols_score/heatmap_bw_org'
    heatmap_lr_dir = '/home/dhaval/Templates/Sahil/OLS_Lib/heat_maps_ols_score/heatmap_bw_wp'
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    thresholds = [0.9]
    
    df = ols_score(image_path, heatmap_hr_dir, heatmap_lr_dir, device, thresholds=thresholds)
    print(df.head())